#ifndef DOCUMENT_STUDIO_H
#define DOCUMENT_STUDIO_H

#include "file/file_aes.h"
#include "file/file_xml.h"
#include "interface/table_structure.h"
#include "channel/channel_bridge.h"
#include <unordered_map>

class EntryDevice;
class EntryTableRecord;
class EntryTable;
class EntryTableField;
class EntryTableList;
class EntryHandling;
class EntrySignalProcessing;
class EntryEventChains;
class EntryIOModule;
class EntryAccessControl2;
class EntryEnergieSchemes;

class DocumentStudio
{
public:
    DocumentStudio( const std::string &studioVersion, const std::string &libraryVersion );
    ~DocumentStudio();

// Flags fuer den Dokumentzustand.
public:
    enum class ResultCodes
    {
        Success,
        FileContentCorrupted,
        OpenUnsupportedFileType,
        OpenUnsupportedFileVersion,
        SetupOpeningFailed,
        FileReadingError,
        FileLoadingError,
        BackupCreationFailed,
        TooMuchDataError,
        FileAES_openFailed,
        FileAES_readFailed,
        FileAES_checksumError,
        FileAES_contentChecksumError,
        FileAES_versionTooNew,
        FileAES_recommendedVersionError,
        FileAES_DFComDllToOld,
        FileAES_wrongFormat,
        FileAES_internalError,
        FileXML_internalError,
        FileXML_structureError,
    };

    struct Result
    {
        Result()
        {
        }

        Result( const Result &rhs )
        {
            *this = rhs;
        }

        Result( ResultCodes resCode )
            : code( resCode )
        {
        }

        Result( ResultCodes resCode, const std::vector<std::string> &data )
            : code( resCode )
            , additionalData( data )
        {
        }

        operator bool() const
        {
            return code == ResultCodes::Success;
        }

        Result &operator=( const Result &rhs )
        {
            if ( this != &rhs )
            {
                code = rhs.code;
                additionalData.clear();
                additionalData = rhs.additionalData;
            }
            return *this;
        }

        ResultCodes code{ ResultCodes::Success };
        std::vector<std::string> additionalData{};
    };

    bool isAccessControl();
    bool isAccessControlExtension();

// Dateiname
public:
    const std::string &filePath()
    {
        return m_filePath;
    }

    void setFilePath( const std::string &fileName )
    {
        m_filePath = fileName;
    }

private:
    std::string m_filePath;

// StudioFile
public:
    FileAES &fileAES()
    {
        return m_fileAES;
    }
    FileXML &fileXML()
    {
        return m_fileXML;
    }
    FileType currentFileType()
    {
        return m_currentFileType;
    }
    void setFileType( FileType type )
    {
        m_currentFileType = type;
    }

    tinyxml2::XMLElement *xmlParentNode()
    {
        return m_parentNodeXML;
    }
    tinyxml2::XMLElement *setXmlParentNode( tinyxml2::XMLElement *newParent )
    {
        auto oldParent = m_parentNodeXML;
        m_parentNodeXML = newParent;
        return oldParent;
    }

    void setVersions( const std::string &studioVersion, const std::string &libraryVersion );
    void setVersions( uint32_t studioVersion, uint32_t libraryVersion );

    std::string studioVersion();              ///< Studioversion mit der die Datei zuletzt abgespeichert wurde.
    std::string minStudioVersion();           ///< Mindestversion des einzusetzenden Studios, um alle enthaltenen Daten zu unterstuetzen.
    std::string shouldStudioVersion();        ///< Sollteversion des einzusetzenden Studios, um alle enthaltenen Daten zu unterstuetzen.
    std::string minLibraryVersion();          ///< Mindestversion der zu verwendenden DFComDLL fuer die korrekte Uebertragung der Setupdaten.
    std::string shouldLibraryVersion();       ///< Sollteversion der zu verwendenden DFComDLL fuer die korrekte Uebertragung der Setupdaten.
    std::string minFirmware1Version();        ///< Mindestversion der zu verwendenden Firmware 04.01.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    std::string shouldFirmware1Version();     ///< Sollteversion der zu verwendenden Firmware 04.01.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    bool supportFirmware1Version();       ///< Wird die Version 04.01.xx der Firmware unterstützt.
    std::string minFirmware2Version();        ///< Mindestversion der zu verwendenden Firmware 04.02.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    std::string shouldFirmware2Version();     ///< Sollteversion der zu verwendenden Firmware 04.02.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    bool supportFirmware2Version();       ///< Wird die Version 04.02.xx der Firmware unterstützt.
    std::string minFirmware3Version();        ///< Mindestversion der zu verwendenden Firmware 04.03.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    std::string shouldFirmware3Version();     ///< Sollteversion der zu verwendenden Firmware 04.03.xx.xx, um alle gewaehlten Einstellungen zu unterstuetzen.
    bool supportFirmware3Version();       ///< Wird die Version 04.03.xx der Firmware unterstützt.

    void SetDocumentDefault( int32_t flags = 0 );

    std::string fileNameList( int32_t num );
    void setFileNameList( int32_t num, const std::string &fileName );

    std::string fileNameListAccessControl( int32_t num );
    void setFileNameListAccessControl( int32_t num, const std::string &fileName );

    std::string fileNameExtended( int32_t num );
    void setFileNameExtended( int32_t num, const std::string &fileName );

private:
    FileAES m_fileAES;
    FileXML m_fileXML;
    std::string versionFormat( uint32_t version );
    tinyxml2::XMLElement *m_parentNodeXML{};
    FileType m_currentFileType{ FileType::AES_AEL };

public:
    const std::string &errorText()
    {
        return m_errorText;
    }

    void setErrorText( const std::string &errorText )
    {
        m_errorText = errorText;
    }

private:
    std::string m_errorText;

public:
    std::list<std::string> globalVariablesName();
    std::unordered_map<std::string, std::string> globalVariablesValueDefault();

    // Dokumenttree <-> Dokumentfile
public:
    uint32_t m_flashSetupSize;
    Result MkSetupImage( EntryDevice &root, uint32_t &setupSize );
    Result BuildRootFromImage( SetupType setupType );
    std::string setupTypeName( SetupType setupType )
    {
        return SetupTypeName.at( setupType );
    }

private:
    EntryDevice *m_root;
    void SetRootDefault( SetupType setupType );

public:
    Result documentNew( SetupType setupType );
    Result documentOpen( const std::string &filePath, bool ignoreStudioVersion,
                         std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup = nullptr,
                         std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo = nullptr );
    bool documentSave();
    bool documentSaveAs( FileType fileType, const std::string &filePath );

private:
    Result documentOpen_AES( const std::string &filePath, bool ignoreStudioVersion,
                             std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup,
                             std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo );
    Result documentOpen_XML( const std::string &filePath, bool ignoreStudioVersion,
                             std::function<bool( const std::string & filePath, const std::string & backupFilePath )> shouldCreateBackup,
                             std::function<bool( const std::string & filePath, const SoftwarebuildVersion & version, const std::string & component )> downgradeInfo );



    // Generierte Funktionen für die Meldungstabellen
public:
    int32_t SaveCheck();

    // Implementierung
public:
    EntryDevice &root();
    EntryTableRecord &rootTableRecord();
    EntryTableList &rootTableList();
    EntryHandling &rootHandling();
    EntrySignalProcessing &rootSignalProcessing();
    EntryEventChains &rootEventChains();
    EntryIOModule &rootIOModule();
    EntryAccessControl2 &rootAccessControl();
    EntryEnergieSchemes &rootEnergieSchemes();

public:
    bool IsEmpty();
    uint32_t GetExtendedFlashSize();
    uint32_t GetMDEEKFlashSize();
    uint32_t GetEntranceEKFlashSize();
    uint32_t GetEntrance2EKFlashSize();
    uint32_t GetFlashSetupSize();
    uint32_t GetFlashSetupSize4_2();

    int32_t IsFunctionAvailable();
    int32_t IsCounterAvailable();
    int32_t IsAnalogValueAvailable();
    int32_t IsAnalogStatusAvailable();
    int32_t IsTimerInputChainAvailable();
    int32_t IsSystemMessageAvailable();

    void GetListName( uint8_t ucNum, char *szBuf );
    std::string listName( int32_t num );

    void GetAccessControlListName( uint8_t ucNum, char *szBuf );
    std::string listNameAccessControl( int32_t num );

    int32_t globalVariableIndex( uint8_t ucDefNum );

    uint8_t globalVariableCount();
    std::string globalVariableName( uint8_t num );

    const char *setupPassword();

protected:
    char *m_setupPassword[17]; // Passwort fuer Kommunikationsschutz

public: // Optionen fuer Einspielen des Dokumentes
    uint8_t m_flagsLoadSetup; // Bit 1: Nach einspielen, die nicht vorbelegten GV's loeschen.
                                    // Bit 2: Nach einspielen, die Systemvariablen der digitalen Eingaenge loeschen.

    // Datensatz und Listenbeschreibungen
public:
    uint8_t GetRecordTableCount();
    EntryTable *GetRecordTable( uint8_t num );
    EntryTableField *GetRecordTableField( uint8_t numTable, uint8_t numField );

    uint8_t GetListTableCount();
    EntryTable *GetListTable( uint8_t num );
    EntryTableField *GetListTableField( uint8_t numTable, uint8_t numField );

    uint8_t GetAccessControlTableCount();
    EntryTable *GetAccessControlTable( uint8_t num );
    EntryTableField *GetAccessControlTableField( uint8_t numTable, uint8_t numField );

    // TableType
#define TT_RECORD         0
#define TT_LIST           1
#define TT_ACCESS_CONTROL 2

public:
    int32_t GetTableCount( uint8_t typeTable );
    int32_t GetTable( uint8_t num, StructTable &table, uint8_t typeTable );
    int32_t GetTableField( uint8_t numTable, uint8_t numField, StructTableField &field, uint8_t typeTable );
};

class RecordStructure : public ITableStructure
{
public:
    RecordStructure( DocumentStudio *doc )
    {
        m_document = doc;
    }

public:
    // Aufbau der Tabelle ermitteln.
    int32_t count()
    {
        return m_document->GetTableCount( TT_RECORD );
    }

    int32_t table( uint8_t num, StructTable &table )
    {
        return m_document->GetTable( num, table, TT_RECORD );
    }

    int32_t field( uint8_t numTable, uint8_t numField, StructTableField &field )
    {
        return m_document->GetTableField( numTable, numField, field, TT_RECORD );
    }

private:
    DocumentStudio *m_document;
};

class ListStructure : public ITableStructure
{
public:
    ListStructure( DocumentStudio *doc )
    {
        m_document = doc;
    }

public:
    // Aufbau der Tabelle ermitteln.
    int32_t count()
    {
        return m_document->GetTableCount( TT_LIST );
    }

    int32_t table( uint8_t num, StructTable &table )
    {
        return m_document->GetTable( num, table, TT_LIST );
    }

    int32_t field( uint8_t numTable, uint8_t numField, StructTableField &field )
    {
        return m_document->GetTableField( numTable, numField, field, TT_LIST );
    }

private:
    DocumentStudio *m_document;
};

class ListAccessControlStructure : public ITableStructure
{
public:
    ListAccessControlStructure( DocumentStudio *doc )
    {
        m_document = doc;
    }

public:
    // Aufbau der Tabelle ermitteln.
    int32_t count()
    {
        return m_document->GetTableCount( TT_ACCESS_CONTROL );
    }

    int32_t table( uint8_t num, StructTable &table )
    {
        return m_document->GetTable( num, table, TT_ACCESS_CONTROL );
    }

    int32_t field( uint8_t numTable, uint8_t numField, StructTableField &field )
    {
        return m_document->GetTableField( numTable, numField, field, TT_ACCESS_CONTROL );
    }

private:
    DocumentStudio *m_document;
};

#endif // DOCUMENT_STUDIO_H
